#include "tiffread.h"
#include "aliases.h"
#include "endianness.h"
#include "image.h"
#include "mem_arena.h"
#include <math.h>
#include <netinet/in.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

#define TIFF_MAGIC 0x002a

#define TIFF_FILENAME_MIN_LENGTH 6
#define TIFF_EXT_MAX_LENGTH 5
#define IS_TIFF_EXTENSION(EXT)                                                 \
  (strncmp(EXT, ".tif", 4) == 0 || strncmp(EXT, ".tiff", 5) == 0)

#define NULL_TIFF_HEADER ((TiffHdr){0})
#define IS_NULL_HEADER(HDR) (HDR.order == 0)

#define NULL_TIFF_IFD ((TiffIFD){0})
#define IS_NULL_IFD(IFD) (IFD.count == 0)

#define NULL_TIFF_IMAGE ((TiffImage){0})
#define IS_NULL_IMAGE(IMG) (IMG.type == TIFF_IMAGE_TYPE_INVALID)

#define INVALID_SAMPLE_COUNT 0
#define INVALID_ROWS_PER_STRIP 0
#define INVALID_STRIP_COUNT 0

#define INVALID_ALPHA_OFFSET -1

TiffHdr read_tiff_header(FILE *fp);
TiffIFD read_ifd(FILE *fp, const TiffHdr *header, u32 offset, Arena *arena);
TiffImage read_fields(FILE *fp, const TiffHdr *header, const TiffIFD *ifd);
bool read_strip_data(FILE *fp, const TiffImage *img, TiffStrip *strips);
void read_strips(FILE *fp, const TiffImage *img, const TiffStrip *strips,
                 Pixel *buf);
bool read_field(const TiffField *field, TiffImage *img,
                u32 *image_type_identifier);
bool validate_image_type(const TiffImage *img,
                         const u32 *image_type_identifier);
void read_from_file_with_offset(FILE *fp, void *dst, u64 count, u64 offset);

Image *read_baseline_tiff(const char *file, Arena *arena) {
  Image *img_out = NULL;

  if (!file) {
    goto READ_BASELINE_RETURN_IMG;
  }

  u64 name_length = strlen(file);
  if (name_length < TIFF_FILENAME_MIN_LENGTH) {
    goto READ_BASELINE_RETURN_IMG;
  }

  const char *ext = NULL;
  for (u32 i = name_length - 1; i >= name_length - TIFF_FILENAME_MIN_LENGTH;
       --i) {
    if (file[i] == '.') {
      ext = &(file[i]);
      break;
    }
  }

  if (!ext || !IS_TIFF_EXTENSION(ext)) {
    goto READ_BASELINE_RETURN_IMG;
  }

  FILE *fp = fopen(file, "rb");
  if (!fp) {
    goto READ_BASELINE_RETURN_IMG;
  }

  TiffHdr header = read_tiff_header(fp);
  if (IS_NULL_HEADER(header)) {
    goto READ_BASELINE_FILE_CLEANUP;
  }

  TiffIFD ifd = read_ifd(fp, &header, header.first_ifd_offset, arena);

  TiffImage img = read_fields(fp, &header, &ifd);
  u64 img_byte_count = sizeof(Pixel) * img.image_width * img.image_length;

  Arena *temp = NULL;
  u64 temp_size = 1 * 1024 * 1024 + img_byte_count;
  if (!wapp_mem_arena_init(&temp, temp_size)) {
    goto READ_BASELINE_FILE_CLEANUP;
  }

  TiffStrip *strips =
      wapp_mem_arena_alloc(temp, sizeof(TiffStrip) * img.strip_count);
  if (!strips) {
    goto READ_BASELINE_DESTROY_ARENA;
  }

  if (!read_strip_data(fp, &img, strips)) {
    goto READ_BASELINE_DESTROY_ARENA;
  }

  for (u32 i = 0; i < img.strip_count; ++i) {
    printf("%u, %u\n", strips[i].offset, strips[i].byte_count);
  }

  Pixel *buf = wapp_mem_arena_alloc(temp, img_byte_count);
  if (!buf) {
    goto READ_BASELINE_DESTROY_ARENA;
  }

  read_strips(fp, &img, strips, buf);

  img_out = create_image(img.image_width, img.image_length, buf, arena);

READ_BASELINE_DESTROY_ARENA:
  wapp_mem_arena_destroy(&temp);

READ_BASELINE_FILE_CLEANUP:
  fclose(fp);

READ_BASELINE_RETURN_IMG:
  return img_out;
}

TiffHdr read_tiff_header(FILE *fp) {
  if (!fp) {
    return NULL_TIFF_HEADER;
  }

  TiffHdr header = NULL_TIFF_HEADER;
  read_from_file_with_offset(fp, &header, sizeof(TiffHdr), 0);

  switch (header.order) {
  case TIFF_ORDER_LITTLE_ENDIAN:
    if (IS_BIG_ENDIAN) {
      header.magic = htons(header.magic);
      header.first_ifd_offset = htonl(header.first_ifd_offset);
    }

    break;
  case TIFF_ORDER_BIG_ENDIAN:
    if (IS_LITTLE_ENDIAN) {
      header.magic = ntohs(header.magic);
      header.first_ifd_offset = ntohl(header.first_ifd_offset);
    }

    break;
  default:
    return NULL_TIFF_HEADER;
  }

  if (header.magic != TIFF_MAGIC) {
    return NULL_TIFF_HEADER;
  }

  return header;
}

TiffIFD read_ifd(FILE *fp, const TiffHdr *header, u32 offset, Arena *arena) {
  if (!fp || !header) {
    return NULL_TIFF_IFD;
  }

  TiffIFD ifd = NULL_TIFF_IFD;
  read_from_file_with_offset(fp, &(ifd.count), sizeof(ifd.count), offset);

  switch (header->order) {
  case TIFF_ORDER_LITTLE_ENDIAN:
    if (IS_BIG_ENDIAN) {
      ifd.count = htons(ifd.count);
    }

    break;
  case TIFF_ORDER_BIG_ENDIAN:
    if (IS_LITTLE_ENDIAN) {
      ifd.count = ntohs(ifd.count);
    }

    break;
  }

  u64 field_byte_count = sizeof(TiffField) * ifd.count;
  ifd.fields = (TiffField *)wapp_mem_arena_alloc(arena, field_byte_count);
  if (!(ifd.fields)) {
    return NULL_TIFF_IFD;
  }

  fread(ifd.fields, field_byte_count, 1, fp);
  fread(&(ifd.next_ifd), sizeof(ifd.next_ifd), 1, fp);

  return ifd;
}

TiffImage read_fields(FILE *fp, const TiffHdr *header, const TiffIFD *ifd) {
  TiffImage img_out = NULL_TIFF_IMAGE;
  u32 image_type_identifier = TIFF_IMAGE_TYPE_BILEVEL;

  for (u64 i = 0; i < ifd->count; ++i) {
    TiffField *field = &(ifd->fields[i]);

    switch (header->order) {
    case TIFF_ORDER_BIG_ENDIAN:
      if (IS_LITTLE_ENDIAN) {
        field->tag = ntohs(field->tag);
        field->type = ntohs(field->type);
        field->count = ntohl(field->count);

        if (field_types[field->type].byte_count > 2 ||
            field_types[field->type].byte_count * field->count > 4) {
          field->value_offset = ntohl(field->value_offset);
        } else {
          field->value_offset = ntohs(field->value_offset);
        }
      }
      break;
    case TIFF_ORDER_LITTLE_ENDIAN:
      if (IS_BIG_ENDIAN) {
        field->tag = htons(field->tag);
        field->type = htons(field->type);
        field->count = htonl(field->count);

        if (field_types[field->type].byte_count > 2 ||
            field_types[field->type].byte_count * field->count > 4) {
          field->value_offset = htonl(field->value_offset);
        } else {
          field->value_offset = htons(field->value_offset);
        }
      }
      break;
    }

    if (!read_field(field, &img_out, &image_type_identifier)) {
      img_out = NULL_TIFF_IMAGE;
      goto READ_FIELDS_RETURN_IMAGE;
    }
  }

  if (!validate_image_type(&img_out, &image_type_identifier)) {
    img_out = NULL_TIFF_IMAGE;
    goto READ_FIELDS_RETURN_IMAGE;
  }

  img_out.type = image_type_identifier;

#if 1
  // clang-format off
  printf("SIZE (width, height)                          :   %u, %u\n", img_out.image_width, img_out.image_length);
  printf("SAMPLES (bits per sample, count, is offset)   :   %u, %u, %u\n", img_out.bits_per_sample, img_out.sample_count, img_out.bits_per_sample_offset);
  printf("EXTRA SAMPLES (samples, count, is offset)     :   %u, %u, %u\n", img_out.extra_samples, img_out.extra_samples_count, img_out.extra_samples_offset);
  printf("SAMPLES PER PIXEL                             :   %u\n", img_out.sample_count);
  printf("PHOTOMETRIC INTERPRETATION                    :   %u\n", img_out.photometric_interpretation);
  printf("ROWS PER STRIP (rows, strip count)            :   %u, %u\n", img_out.rows_per_strip, img_out.strip_count);
  printf("STRIP OFFSETS (offsets, is offset)            :   %u, %u\n", img_out.strip_offsets.long_val, img_out.strip_offsets_offset);
  printf("STRIP BYTES (byte count, is offset)           :   %u, %u\n", img_out.strip_byte_counts.long_val, img_out.strip_byte_counts_offset);
  // clang-format on
#endif

READ_FIELDS_RETURN_IMAGE:
  return img_out;
}

bool read_strip_data(FILE *fp, const TiffImage *img, TiffStrip *strips) {
  if (!fp || !img || !strips) {
    return false;
  }

  u32 offsets_total_bytes =
      img->strip_count * img->strip_offsets_type_byte_count;
  u32 byte_count_total_bytes =
      img->strip_count * img->strip_byte_count_type_byte_count;

  if ((!(img->strip_offsets_offset) && offsets_total_bytes > sizeof(u32)) ||
      (!(img->strip_byte_counts_offset) &&
       byte_count_total_bytes > sizeof(u32))) {
    return false;
  }

  for (u32 i = 0; i < img->strip_count; ++i) {
    TiffStrip *strip = &(strips[i]);

    u16 offset_size = img->strip_offsets_type_byte_count;
    u16 counts_size = img->strip_byte_count_type_byte_count;

    if (offsets_total_bytes > sizeof(u32)) {
      u32 offset = img->strip_offsets.long_val + offset_size * i;
      read_from_file_with_offset(fp, &(strip->offset), offset_size, offset);
    } else {
      memcpy(&(strip->offset), &(img->strip_offsets.long_val), offset_size);
    }

    if (byte_count_total_bytes > sizeof(u32)) {
      u32 offset = img->strip_byte_counts.long_val + counts_size * i;
      read_from_file_with_offset(fp, &(strip->byte_count), counts_size, offset);
    } else {
      memcpy(&(strip->byte_count), &(img->strip_byte_counts.long_val),
             counts_size);
    }
  }

  return true;
}

void read_strips(FILE *fp, const TiffImage *img, const TiffStrip *strips,
                 Pixel *buf) {
  for (u64 i = 0; i < img->strip_count; ++i) {
    const TiffStrip *strip = &(strips[i]);

    for (u64 j = 0; j < img->image_width * img->image_length; ++j) {
      Pixel *p = &(buf[j]);

      read_from_file_with_offset(fp, p, 3, strip->offset + j * 3);

      p->a = 255;
    }
  }
}

bool read_field(const TiffField *field, TiffImage *img,
                u32 *image_type_identifier) {
  switch (field->tag) {
  case TIFF_PUBLIC_TAG_IMAGE_WIDTH:
    if (field->count > 1) {
      return false;
    }

    img->image_width = field->value_offset;

    break;
  case TIFF_PUBLIC_TAG_IMAGE_LENGTH:
    if (field->count > 1) {
      return false;
    }

    img->image_length = field->value_offset;

    break;
  case TIFF_PUBLIC_TAG_BITS_PER_SAMPLE:
    *image_type_identifier |= field->tag;

    if (img->sample_count == INVALID_SAMPLE_COUNT) {
      img->sample_count = field->count;
    } else if (img->sample_count != field->count) {
      return false;
    }

    img->bits_per_sample = field->value_offset;

    img->bits_per_sample_offset =
        img->sample_count * field_types[field->type].byte_count > 4;

    break;
  case TIFF_PUBLIC_TAG_COMPRESSION:
    if (field->count > 1 ||
        field->value_offset != TIFF_COMPRESSION_UNCOMPRESSED) {
      return false;
    }

    img->compression = field->value_offset;

    break;
  case TIFF_PUBLIC_TAG_PHOTOMETRIC_INTERPRETATION:
    if (field->count > 1) {
      return false;
    }

    img->photometric_interpretation = field->value_offset;

    break;
  case TIFF_PUBLIC_TAG_STRIP_OFFSETS:
    img->strip_offsets_type_byte_count = field_types[field->type].byte_count;

    if (img->strip_offsets_type_byte_count == sizeof(u16)) {
      img->strip_offsets.short_val = field->value_offset;
    } else if (img->strip_offsets_type_byte_count == sizeof(u32)) {
      img->strip_offsets.long_val = field->value_offset;
    }

    if (img->strip_count == INVALID_STRIP_COUNT) {
      img->strip_count = field->count;
    } else if (img->strip_count != field->count) {
      return false;
    }

    img->strip_offsets_offset =
        img->strip_count * img->strip_offsets_type_byte_count > 4;

    break;
  case TIFF_PUBLIC_TAG_SAMPLES_PER_PIXEL:
    *image_type_identifier |= field->tag;

    if (field->count > 1) {
      return false;
    }

    img->sample_count = field->value_offset;

    break;
  case TIFF_PUBLIC_TAG_ROWS_PER_STRIP:
    if (field->count > 1) {
      return false;
    }

    img->rows_per_strip = field->value_offset;

    if (img->image_length > 0) {
      img->strip_count =
          floor((f64)(img->image_length + img->rows_per_strip - 1) /
                img->rows_per_strip);
    } else {
      img->strip_count = INVALID_STRIP_COUNT;
    }

    break;
  case TIFF_PUBLIC_TAG_STRIP_BYTE_COUNTS:
    img->strip_byte_count_type_byte_count = field_types[field->type].byte_count;

    if (img->strip_byte_count_type_byte_count == sizeof(u16)) {
      img->strip_byte_counts.short_val = field->value_offset;
    } else if (img->strip_byte_count_type_byte_count == sizeof(u32)) {
      img->strip_byte_counts.long_val = field->value_offset;
    }

    if (img->strip_count == INVALID_STRIP_COUNT) {
      img->strip_count = field->count;
    } else if (img->strip_count != field->count) {
      return false;
    }

    img->strip_byte_counts_offset =
        img->strip_count * img->strip_byte_count_type_byte_count > 4;

    break;
  case TIFF_PUBLIC_TAG_COLOR_MAP:
    *image_type_identifier |= field->tag;
    break;
  case TIFF_PUBLIC_TAG_EXTRA_SAMPLES:
    img->extra_samples_count = field->count;

    img->extra_samples = field->value_offset;

    img->extra_samples_offset =
        img->extra_samples_count * field_types[field->type].byte_count > 4;

    break;
  case TIFF_PUBLIC_TAG_PLANAR_CONFIGURATION:
    if (field->count > 1 || field->value_offset != TIFF_PLANAR_CONFIG_CHUNKY) {
      return false;
    }

    break;
  }

  return true;
}

bool validate_image_type(const TiffImage *img,
                         const u32 *image_type_identifier) {
  if (*image_type_identifier < TIFF_IMAGE_TYPE_BILEVEL ||
      *image_type_identifier > TIFF_IMAGE_TYPE_RGB) {
    return false;
  }

  switch (*image_type_identifier) {
  case TIFF_IMAGE_TYPE_BILEVEL:
  case TIFF_IMAGE_TYPE_GRAYSCALE:
    if (img->photometric_interpretation >
        TIFF_PHOTOMETRIC_INTERPRETATION_BLACK_IS_ZERO) {
      return false;
    }

    break;
  case TIFF_IMAGE_TYPE_PALETTE:
    if (img->photometric_interpretation !=
        TIFF_PHOTOMETRIC_INTERPRETATION_RGB_PALETTE) {
      return false;
    }

    break;
  case TIFF_IMAGE_TYPE_RGB:
    if (img->photometric_interpretation !=
        TIFF_PHOTOMETRIC_INTERPRETATION_RGB) {
      return false;
    }

    break;
  default:
    return false;
  }

  return true;
}

void read_from_file_with_offset(FILE *fp, void *dst, u64 count, u64 offset) {
  if (!fp || !dst) {
    return;
  }

  fseek(fp, offset, SEEK_SET);

  fread(dst, count, 1, fp);
}
